"""Verify the calls made during rehost."""


import os
import mock
import pytest
import tempfile

from pypicloud_tools import rehost
from pypicloud_tools import Settings
from pypicloud_tools import S3Config
from pypicloud_tools import PyPIConfig


def test_cleanup_tempdir():
    """Ensure the tempdir used is removed afterwards."""

    starting = os.listdir(tempfile.tempdir)
    fake_s3 = S3Config("fake_bucket", "fake_access", "fake_secret", "fake_acl")
    fake_pypi = PyPIConfig("fake_server", "fake_user", "fake_passwd")
    fake_settings = Settings(fake_s3, fake_pypi, ["requests"])

    with mock.patch.object(rehost.pip, "main") as patched_pip:
        with mock.patch.object(rehost, "upload_files") as patched_upload:
            with mock.patch.object(rehost, "get_bucket_conn") as patched_conn:
                with mock.patch.object(rehost, "get_settings", return_value=fake_settings):
                    rehost.main()

    assert len(starting) == len(os.listdir(tempfile.tempdir))

    pip_args = patched_pip.mock_calls[0][1][0]
    assert pip_args[0] == "install"
    assert pip_args[1] == "--download"
    assert pip_args[2].startswith(tempfile.tempdir)
    assert pip_args[3] == "requests"

    # items should be empty here because we've mocked downloading anything
    patched_upload.assert_called_once_with(
        Settings(fake_s3, fake_pypi, []),
        patched_conn(),
    )


if __name__ == "__main__":
    pytest.main(["-v", "-rx", "--pdb", __file__])



"""Ensure the pypicloud_tools lister functions work as expected."""


import mock
import pytest

import pypicloud_tools
from pypicloud_tools import lister


def test_main():
    """Mock all calls and verify the main lister application flow."""

    bucket = mock.Mock()
    settings = mock.Mock()
    settings.items = ["faked"]

    with mock.patch.object(lister, "get_settings", return_value=settings) as settings_patch:
        with mock.patch.object(lister, "get_bucket_conn", return_value=bucket) as get_bucket_patch:
            with mock.patch.object(lister, "list_package") as lister_patch:
                with mock.patch.object(lister, "parse_package", return_value=(1, 2)) as parse_patch:
                    lister.main()

    settings_patch.assert_called_once_with(listing=True)
    get_bucket_patch.assert_called_once_with(settings.s3)
    lister_patch.assert_called_once_with(bucket, 1, 2)
    parse_patch.assert_called_once_with("faked")


def test_main_buries_errors(capfd):
    """Any Exception thrown in list_package should be handled."""

    bucket = mock.Mock()
    settings = mock.Mock()
    settings.items = ["faked"]

    with mock.patch.object(lister, "get_settings", return_value=settings) as settings_patch:
        with mock.patch.object(lister, "get_bucket_conn", return_value=bucket) as get_bucket_patch:
            with mock.patch.object(lister, "list_package", side_effect=IOError) as lister_patch:
                with mock.patch.object(lister, "parse_package", return_value=(1, 2)) as parse_patch:
                    lister.main()

    settings_patch.assert_called_once_with(listing=True)
    get_bucket_patch.assert_called_once_with(settings.s3)
    lister_patch.assert_called_once_with(bucket, 1, 2)
    parse_patch.assert_called_once_with("faked")

    out, err = capfd.readouterr()
    assert not out
    assert "Error listing faked: " in err


def test_print_versioned__in_order(capfd):
    releases = [
        "some_thing-0.0.1",
        "some_thing-1.2.4",
        "some_thing-1.5.2.dev1",
        "some_thing-2.4.3",
        "some_thing-1.5.2",
        "some_thing-1.7.5",
    ]
    expected = [
        "some_thing=2.4.3",
        "some_thing=1.7.5",
        "some_thing=1.5.2",
        "some_thing=1.5.2.dev1",
        "some_thing=1.2.4",
        "some_thing=0.0.1",
    ]
    lister.print_versioned(releases, "some_thing")
    out, err = capfd.readouterr()
    assert not err
    assert "\n".join(expected) in out


def test_list_packages__all_packages(capfd, bucket_and_keys):
    """If no package is provided, all base names should be listed."""

    lister.list_package(bucket_and_keys[0], None)
    out, err = capfd.readouterr()
    assert not err
    assert "error_pkg\npackage-one\npackage_two" in out


def test_list_packages__specific(capfd, bucket_and_keys):
    """Verify the exact release is listed if requested."""

    with mock.patch.object(lister, "print_versioned") as patched_print:
        lister.list_package(bucket_and_keys[0], "package_two", "0.0.1")

    patched_print.assert_called_once_with(
        [
            "package_two-0.0.1.dev1-py2.py3-none-any.whl",
            "package_two-0.0.1.dev2-py2.py3-none-any.whl",
            "package_two-0.0.1-py2.py3-none-any.whl",
            "package_two-0.0.1.tar.gz",
            "package_two-0.0.1-py2.7.egg",
        ],
        "package_two",
    )


if __name__ == "__main__":
    pytest.main(["-v", "-rx", "--pdb", __file__])



"""Ensure the download functions work as expected."""


import sys
import mock
import pytest

if sys.version_info.major == 2:
    import __builtin__ as builtins
else:
    import builtins

import pypicloud_tools
from pypicloud_tools import download


@pytest.fixture
def argv_cleanup(request, autouse=True, scope="function"):
    """Adds a finalizer to clean up sys.argv in download."""

    sys._argv = sys.argv
    request.addfinalizer(argv_cleaner)


def argv_cleaner():
    """Reset download.sys.argv."""

    sys.argv = sys._argv
    del sys._argv


@pytest.fixture
def isatty_cleanup(request):
    """Adds a finalizer to clean up sys.stdout.isatty in download."""

    sys.stdout._isatty = sys.stdout.isatty
    request.addfinalizer(isatty_cleaner)


def isatty_cleaner():
    """Reset download.sys.stdout.isatty."""

    sys.stdout._isatty = sys.stdout.isatty
    del sys.stdout._isatty


def test_main():
    """Mocks everything to ensure the main entry point program flow."""

    bucket = mock.Mock()
    settings = mock.Mock()
    settings.items = ["faked"]

    with mock.patch.object(download, "get_settings", return_value=settings) as settings_patch:
        with mock.patch.object(download, "get_bucket_conn", return_value=bucket) as get_bucket_patch:
            with mock.patch.object(download, "download_package") as download_patch:
                with mock.patch.object(download, "parse_package", return_value=(1, 2)) as parse_patch:
                    download.main()

    settings_patch.assert_called_once_with(download=True)
    get_bucket_patch.assert_called_once_with(settings.s3)
    download_patch.assert_called_once_with(bucket, 1, 2)
    parse_patch.assert_called_once_with("faked")


def test_main_buries_errors(capfd):
    """Any Exception thrown in download_package should be handled."""

    bucket = mock.Mock()
    settings = mock.Mock()
    settings.items = ["faked"]

    with mock.patch.object(download, "get_settings", return_value=settings) as settings_patch:
        with mock.patch.object(download, "get_bucket_conn", return_value=bucket) as get_bucket_patch:
            with mock.patch.object(download, "download_package", side_effect=IOError) as download_patch:
                with mock.patch.object(download, "parse_package", return_value=(1, 2)) as parse_patch:
                    download.main()

    settings_patch.assert_called_once_with(download=True)
    get_bucket_patch.assert_called_once_with(settings.s3)
    download_patch.assert_called_once_with(bucket, 1, 2)
    parse_patch.assert_called_once_with("faked")

    out, err = capfd.readouterr()
    assert not out
    assert "Error downloading faked: " in err


def test_download_package__specific(bucket_and_keys):
    """Verify the calls made to succesfully download a specific package."""

    bucket, keys = bucket_and_keys

    with mock.patch.object(download, "write_key") as patched_write:
        download.download_package(bucket, "package-one", "1.2.3-alpha1")

    patched_write.assert_called_once_with(keys[0])


def test_download_package__not_found(bucket_and_keys):
    """SystemExit should be raised when the package is not found."""

    with pytest.raises(SystemExit) as exit_error:
        download.download_package(bucket_and_keys[0], "package-unknown")

    assert "Package package-unknown not found" in exit_error.value.args

    with pytest.raises(SystemExit) as specific_error:
        download.download_package(bucket_and_keys[0], "something", "1.2.3")

    assert "Package something=1.2.3 not found" in specific_error.value.args


def test_download_package__perfers_wheels(bucket_and_keys):
    """When multiple versions are available, by default prefer wheels."""

    bucket, keys = bucket_and_keys

    with mock.patch.object(download, "write_key") as patched_write:
        download.download_package(bucket, "package_two", "0.0.1")

    patched_write.assert_called_once_with(keys[6])


def test_download_package__prefer_egg(bucket_and_keys, argv_cleanup):
    """Ensure you can receive an egg if you request one."""

    bucket, keys = bucket_and_keys
    download.sys.argv = ["download", "package_two=0.0.1", "--egg"]

    with mock.patch.object(download, "write_key") as patched_write:
        download.download_package(bucket, "package_two", "0.0.1")

    patched_write.assert_called_once_with(keys[8])


def test_download_package__prefer_src(bucket_and_keys, argv_cleanup):
    """Ensure you can receive a source package if it's requested."""

    bucket, keys = bucket_and_keys
    download.sys.argv = ["download", "package_two=0.0.1", "--src"]

    with mock.patch.object(download, "write_key") as patched_write:
        download.download_package(bucket, "package_two", "0.0.1")

    patched_write.assert_called_once_with(keys[7])


def test_download_package__too_many_packages(bucket_and_keys):
    """If requesting a package with too many options, raise SystemExit."""

    bucket, keys = bucket_and_keys
    with pytest.raises(SystemExit) as exit_error:
        what = download.prefer_wheels(keys, "error_pkg")

    expected = (
        "Found too many results for error_pkg:\n"
        "  error_pkg/error_pkg-2.3.4-py2.py3-none-any.whl\n"
        "  error_pkg/error_pkg-2.3.4-py2-none-any.whl\n"
        "  error_pkg/error_pkg-2.3.4.tar.gz"
    )
    assert expected in exit_error.value.args

    with pytest.raises(SystemExit) as exit_error:
        what = download.prefer_wheels(keys, "error_pkg", "2.3.4")

    expected = (
        "Found too many results for error_pkg=2.3.4:\n"
        "  error_pkg/error_pkg-2.3.4-py2.py3-none-any.whl\n"
        "  error_pkg/error_pkg-2.3.4-py2-none-any.whl\n"
        "  error_pkg/error_pkg-2.3.4.tar.gz"
    )
    assert expected in exit_error.value.args


def test_write_key__to_file(capfd, isatty_cleanup):
    """Verify the calls made to write a S3 key to file."""

    download.sys.stdout.isatty = mock.Mock(return_value=True)
    key = mock.Mock()
    key.name = "mock_pkg/mock_pkg-1.0.2-dev1.tar.gz"

    with mock.patch.object(builtins, "open") as open_patch:
        download.write_key(key)

    key.get_contents_to_file.assert_called_once_with(open_patch().__enter__())
    out, err = capfd.readouterr()
    assert not err
    assert "mock_pkg-1.0.2-dev1.tar.gz" in out


def test_write_key__to_stdout(isatty_cleanup):
    """When sys.stdout is being piped/redicrected, print contents to it."""

    download.sys.stdout.isatty = mock.Mock(return_value=False)
    key = mock.Mock()
    download.write_key(key)
    key.get_contents_to_file.assert_called_once_with(download.sys.stdout)


@pytest.mark.parametrize("flag", ("--url", "--url-only"))
def test_write_key__generate_url(flag, capfd, argv_cleanup):
    """Ensure that we print a url when asked."""

    key = mock.Mock()
    download.sys.argv = ["download", flag]
    download.write_key(key)
    key.generate_url.assert_called_once_with(300)  # 5 minute URLs
    out, err = capfd.readouterr()
    assert not err
    assert str(key.generate_url(300)) in out


if __name__ == "__main__":
    pytest.main(["-v", "-rx", "--pdb", __file__])



"""pypicloud-tools setup.py"""


import io
from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """TestCommand subclass to use pytest with setup.py test."""

    def finalize_options(self):
        """Find our package name and test options to fill out test_args."""

        TestCommand.finalize_options(self)
        self.test_args = ['-v', '-rx', '--cov', 'pypicloud_tools',
            '--cov-report', 'term-missing']
        self.test_suite = True

    def run_tests(self):
        """Taken from http://pytest.org/latest/goodpractises.html."""

        # have to import here, outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        raise SystemExit(errno)


with io.open("README.rst", "r", encoding="utf-8") as openreadme:
    long_description = openreadme.read()


setup(
    name="pypicloud-tools",
    version="0.0.6",
    author="Adam Talsma",
    author_email="se-adam.talsma@ccpgames.com",
    packages=["pypicloud_tools"],
    entry_points={"console_scripts": [
        "upload = pypicloud_tools.upload:main",
        "download = pypicloud_tools.download:main",
        "list = pypicloud_tools.lister:main",
        "rehost = pypicloud_tools.rehost:main",
    ]},
    install_requires=[
        "boto >= 2.38.0",
        "futures >= 2.2.0",
        "futures < 3.0.0",
        "filechunkio >= 1.6",
        "requests >= 2.6.2",
        "setuptools >= 15.0",
    ],
    tests_require=["pytest", "pytest-cov", "mock"],
    cmdclass={"test": PyTest},
    description="Tools to bypass PyPICloud and work with S3 directly",
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Topic :: System :: Software Distribution",
        "Topic :: Utilities",
    ],
)



"""Rehosts packages from PyPI in pypicloud.

Copyright (c) 2015 CCP Games. Released for use under the MIT license.
"""


import os
import pip
import shutil
import tempfile

from . import Settings
from . import get_settings
from . import get_bucket_conn
from .upload import upload_files


class TempDir(object):
    """Context manager for storing the in-transit files in temp storage."""
    def __init__(self):
        self.dir = tempfile.mkdtemp()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        shutil.rmtree(self.dir)


def main():
    """Entry point for rehosting PyPI packages on pypicloud."""

    settings = get_settings(rehost=True)
    bucket = get_bucket_conn(settings.s3)

    with TempDir() as storage:
        for package in settings.items:
            pip.main(["install", "--download", storage.dir, package])

        upload_settings = Settings(
            settings.s3,
            settings.pypi,
            [os.path.join(storage.dir, f) for f in os.listdir(storage.dir)],
        )
        upload_files(upload_settings, bucket)
